include "mjmodel.pxd"
include "mjdata.pxd"
include "mjrender.pxd"
include "mjvisualize.pxd"


cdef extern from "mujoco.h" nogil:
    # macros
    #define mjMARKSTACK   int _mark = d->pstack;
    #define mjFREESTACK   d->pstack = _mark;
    #define mjDISABLED(x) (m->opt.disableflags & (x))
    #define mjENABLED(x)  (m->opt.enableflags & (x))


    # user error and memory handlers
    void  (*mju_user_error)(const char*);
    void  (*mju_user_warning)(const char*);
    void* (*mju_user_malloc)(size_t);
    void  (*mju_user_free)(void*);


    # # callbacks extending computation pipeline
    # mjfGeneric  mjcb_passive;
    # mjfGeneric  mjcb_control;
    # mjfSensor   mjcb_sensor;
    # mjfTime     mjcb_time;
    # mjfAct      mjcb_act_dyn;
    # mjfAct      mjcb_act_gain;
    # mjfAct      mjcb_act_bias;
    # mjfSolImp   mjcb_sol_imp;
    # mjfSolRef   mjcb_sol_ref;
    #
    #
    # # collision function table
    # mjfCollision mjCOLLISIONFUNC[mjNGEOMTYPES][mjNGEOMTYPES];
    #
    #
    # # string names
    const char* mjDISABLESTRING[mjNDISABLE];
    const char* mjENABLESTRING[mjNENABLE];
    const char* mjTIMERSTRING[mjNTIMER];
    const char* mjLABELSTRING[mjNLABEL];
    const char* mjFRAMESTRING[mjNFRAME];
    const char* mjVISSTRING[mjNVISFLAG][3];
    const char* mjRNDSTRING[mjNRNDFLAG][3];


    #---------------------- Activation -----------------------------------------------------

    # activate license, call mju_error on failure; return 1 if ok, 0 if failure
    int mj_activate(const char* filename);

    # deactivate license, free memory
    void mj_deactivate();

    # server: generate certificate question
    void mj_certQuestion(mjtNum question[16]);

    # client: generate certificate answer given question
    void mj_certAnswer(const mjtNum question[16], mjtNum answer[16]);

    # server: check certificate question-answer pair; return 1 if match, 0 if mismatch
    int mj_certCheck(const mjtNum question[16], const mjtNum answer[16]);

    #---------------------- Virtual file system --------------------------------------------

    # Initialize VFS to empty (no deallocation).
    void mj_defaultVFS(mjVFS* vfs);

    # Add file to VFS, return 0: success, 1: full, 2: repeated name, -1: not found on disk.
    int mj_addFileVFS(mjVFS* vfs, const char* directory, const char* filename);

    # Make empty file in VFS, return 0: success, 1: full, 2: repeated name.
    int mj_makeEmptyFileVFS(mjVFS* vfs, const char* filename, int filesize);

    # Return file index in VFS, or -1 if not found in VFS.
    int mj_findFileVFS(const mjVFS* vfs, const char* filename);

    # Delete file from VFS, return 0: success, -1: not found in VFS.
    int mj_deleteFileVFS(mjVFS* vfs, const char* filename);

    # Delete all files from VFS.
    void mj_deleteVFS(mjVFS* vfs);

    #--------------------- Parse and compile ----------------------------------------------

    # Parse XML file in MJCF or URDF format, compile it, return low-level model.
    # If vfs is not NULL, look up files in vfs before reading from disk.
    # If error is not NULL, it must have size error_sz.
    mjModel* mj_loadXML(const char* filename, const mjVFS* vfs, 
                              char* error, int error_sz);

    # Update XML data structures with info from low-level model, save as MJCF.
    # If error is not NULL, it must have size error_sz.
    int mj_saveLastXML(const char* filename, const mjModel* m, 
                             char* error, int error_sz);

    # Free last XML model if loaded. Called internally at each load.
    void mj_freeLastXML();

    # Print internal XML schema as plain text or HTML, with style-padding or &nbsp;.
    int mj_printSchema(const char* filename, char* buffer, int buffer_sz, 
                             int flg_html, int flg_pad);


    #--------------------- Main simulation ------------------------------------------------

    # Advance simulation, use control callback to obtain external force and control.
    void mj_step(const mjModel* m, mjData* d);

    # Advance simulation in two steps: before external force and control is set by user.
    void mj_step1(const mjModel* m, mjData* d);

    # Advance simulation in two steps: after external force and control is set by user.
    void mj_step2(const mjModel* m, mjData* d);

    # Forward dynamics: same as mj_step but do not integrate in time.
    void mj_forward(const mjModel* m, mjData* d);

    # Inverse dynamics: qacc must be set before calling.
    void mj_inverse(const mjModel* m, mjData* d);

    # Forward dynamics with skip; skipstage is mjtStage.
    void mj_forwardSkip(const mjModel* m, mjData* d, 
                              int skipstage, int skipsensorenergy);

    # Inverse dynamics with skip; skipstage is mjtStage.
    void mj_inverseSkip(const mjModel* m, mjData* d, 
                              int skipstage, int skipsensorenergy);


    #--------------------- Initialization -------------------------------------------------

    # Set solver parameters to default values.
    void mj_defaultSolRefImp(mjtNum* solref, mjtNum* solimp);

    # Set physics options to default values.
    void mj_defaultOption(mjOption* opt);

    # Set visual options to default values.
    void mj_defaultVisual(mjVisual* vis);

    # Copy mjModel, allocate new if dest is NULL.
    mjModel* mj_copyModel(mjModel* dest, const mjModel* src);

    # Save model to binary MJB file or memory buffer; buffer has precedence when given.
    void mj_saveModel(const mjModel* m, const char* filename, void* buffer, int buffer_sz);

    # Load model from binary MJB file.
    # If vfs is not NULL, look up file in vfs before reading from disk.
    mjModel* mj_loadModel(const char* filename, mjVFS* vfs);

    # Free memory allocation in model.
    void mj_deleteModel(mjModel* m);

    # Return size of buffer needed to hold model.
    int mj_sizeModel(const mjModel* m);

    # Allocate mjData correponding to given model.
    mjData* mj_makeData(const mjModel* m);

    # Copy mjData.
    mjData* mj_copyData(mjData* dest, const mjModel* m, const mjData* src);

    # Reset data to defaults.
    void mj_resetData(const mjModel* m, mjData* d);

    # Reset data to defaults, fill everything else with debug_value.
    void mj_resetDataDebug(const mjModel* m, mjData* d, unsigned char debug_value);

    # Reset data, set fields from specified keyframe.
    void mj_resetDataKeyframe(const mjModel* m, mjData* d, int key);

    # Allocate array of specified size on mjData stack. Call mju_error on stack overflow.
    mjtNum* mj_stackAlloc(mjData* d, int size);

    # Free memory allocation in mjData.
    void mj_deleteData(mjData* d);

    # Reset all callbacks to NULL pointers (NULL is the default).
    void mj_resetCallbacks();

    # Set constant fields of mjModel, corresponding to qpos0 configuration.
    # The flag flg_actrange currently has no effect.
    void mj_setConst(mjModel* m, mjData* d, int flg_actrange);


    #--------------------- Printing -------------------------------------------------------

    # Print model to text file.
    void mj_printModel(const mjModel* m, const char* filename); 

    # Print data to text file.
    void mj_printData(const mjModel* m, mjData* d, const char* filename); 

    # Print matrix to screen.
    void mju_printMat(const mjtNum* mat, int nr, int nc);

    # Print sparse matrix to screen.
    void mju_printMatSparse(const mjtNum* mat, int nr,
                                  const int* rownnz, const int* rowadr, 
                                  const int* colind);


    #--------------------- Components -----------------------------------------------------

    # Run position-dependent computations.
    void mj_fwdPosition(const mjModel* m, mjData* d);

    # Run velocity-dependent computations.
    void mj_fwdVelocity(const mjModel* m, mjData* d);

    # Compute actuator force qfrc_actuation.
    void mj_fwdActuation(const mjModel* m, mjData* d);

    # Add up all non-constraint forces, compute qacc_unc.
    void mj_fwdAcceleration(const mjModel* m, mjData* d);

    # Run selected constraint solver.
    void mj_fwdConstraint(const mjModel* m, mjData* d);

    # Euler integrator, semi-implicit in velocity.
    void mj_Euler(const mjModel* m, mjData* d);

    # Runge-Kutta explicit order-N integrator.
    void mj_RungeKutta(const mjModel* m, mjData* d, int N);

    # Run position-dependent computations in inverse dynamics.
    void mj_invPosition(const mjModel* m, mjData* d);

    # Run velocity-dependent computations in inverse dynamics.
    void mj_invVelocity(const mjModel* m, mjData* d);

    # Apply the analytical formula for inverse constraint dynamics.
    void mj_invConstraint(const mjModel* m, mjData* d);

    # Compare forward and inverse dynamics, save results in fwdinv.
    void mj_compareFwdInv(const mjModel* m, mjData* d);


    #--------------------- Sub components -------------------------------------------------

    # Evaluate position-dependent sensors.
    void mj_sensorPos(const mjModel* m, mjData* d);

    # Evaluate velocity-dependent sensors.
    void mj_sensorVel(const mjModel* m, mjData* d);

    # Evaluate acceleration and force-dependent sensors.
    void mj_sensorAcc(const mjModel* m, mjData* d);

    # Evaluate position-dependent energy (potential).
    void mj_energyPos(const mjModel* m, mjData* d);

    # Evaluate velocity-dependent energy (kinetic).
    void mj_energyVel(const mjModel* m, mjData* d);

    # Check qpos, reset if any element is too big or nan.
    void mj_checkPos(const mjModel* m, mjData* d);

    # Check qvel, reset if any element is too big or nan.
    void mj_checkVel(const mjModel* m, mjData* d);

    # Check qacc, reset if any element is too big or nan.
    void mj_checkAcc(const mjModel* m, mjData* d);

    # Run forward kinematics.
    void mj_kinematics(const mjModel* m, mjData* d);

    # Map inertias and motion dofs to global frame centered at CoM.
    void mj_comPos(const mjModel* m, mjData* d);

    # Compute camera and light positions and orientations.
    void mj_camlight(const mjModel* m, mjData* d);

    # Compute tendon lengths, velocities and moment arms.
    void mj_tendon(const mjModel* m, mjData* d);

    # Compute actuator transmission lengths and moments.
    void mj_transmission(const mjModel* m, mjData* d);

    # Run composite rigid body inertia algorithm (CRB).
    void mj_crb(const mjModel* m, mjData* d);

    # Compute sparse L'*D*L factorizaton of inertia matrix.
    void mj_factorM(const mjModel* m, mjData* d);

    # Solve linear system M * x = y using factorization:  x = inv(L'*D*L)*y
    void mj_solveM(const mjModel* m, mjData* d, mjtNum* x, const mjtNum* y, int n);

    # Half of linear solve:  x = sqrt(inv(D))*inv(L')*y
    void mj_solveM2(const mjModel* m, mjData* d, mjtNum* x, const mjtNum* y, int n);

    # Compute cvel, cdof_dot.
    void mj_comVel(const mjModel* m, mjData* d);

    # Compute qfrc_passive from spring-dampers, viscosity and density.
    void mj_passive(const mjModel* m, mjData* d);

    # RNE: compute M(qpos)*qacc + C(qpos,qvel); flg_acc=0 removes inertial term.
    void mj_rne(const mjModel* m, mjData* d, int flg_acc, mjtNum* result);

    # RNE with complete data: compute cacc, cfrc_ext, cfrc_int.
    void mj_rnePostConstraint(const mjModel* m, mjData* d);

    # Run collision detection.
    void mj_collision(const mjModel* m, mjData* d);

    # Construct constraints.
    void mj_makeConstraint(const mjModel* m, mjData* d);

    # Compute inverse constaint inertia efc_AR.
    void mj_projectConstraint(const mjModel* m, mjData* d);

    # Compute efc_vel, efc_aref.
    void mj_referenceConstraint(const mjModel* m, mjData* d);

    # Compute efc_state, efc_force, qfrc_constraint, and (optionally) cone Hessians.
    # If cost is not NULL, set *cost = s(jar) where jar = Jac*qacc-aref.
    void mj_constraintUpdate(const mjModel* m, mjData* d, const mjtNum* jar,
                                   mjtNum* cost, int flg_coneHessian);


    #--------------------- Support --------------------------------------------------------

    # Add contact to d->contact list; return 0 if success; 1 if buffer full.
    int mj_addContact(const mjModel* m, mjData* d, const mjContact* con);

    # Determine type of friction cone.
    int mj_isPyramidal(const mjModel* m);

    # Determine type of constraint Jacobian.
    int mj_isSparse(const mjModel* m);

    # Determine type of solver (PGS is dual, CG and Newton are primal).
    int mj_isDual(const mjModel* m);

    # Multiply dense or sparse constraint Jacobian by vector.
    void mj_mulJacVec(const mjModel* m, mjData* d, 
                            mjtNum* res, const mjtNum* vec);

    # Multiply dense or sparse constraint Jacobian transpose by vector.
    void mj_mulJacTVec(const mjModel* m, mjData* d, mjtNum* res, const mjtNum* vec);

    # Compute 3/6-by-nv end-effector Jacobian of global point attached to given body.
    void mj_jac(const mjModel* m, const mjData* d, 
                      mjtNum* jacp, mjtNum* jacr, const mjtNum point[3], int body);

    # Compute body frame end-effector Jacobian.
    void mj_jacBody(const mjModel* m, const mjData* d, 
                          mjtNum* jacp, mjtNum* jacr, int body);

    # Compute body center-of-mass end-effector Jacobian.
    void mj_jacBodyCom(const mjModel* m, const mjData* d, 
                             mjtNum* jacp, mjtNum* jacr, int body);

    # Compute geom end-effector Jacobian.
    void mj_jacGeom(const mjModel* m, const mjData* d, 
                          mjtNum* jacp, mjtNum* jacr, int geom);

    # Compute site end-effector Jacobian.
    void mj_jacSite(const mjModel* m, const mjData* d, 
                          mjtNum* jacp, mjtNum* jacr, int site);

    # Compute translation end-effector Jacobian of point, and rotation Jacobian of axis.
    void mj_jacPointAxis(const mjModel* m, mjData* d, 
                               mjtNum* jacPoint, mjtNum* jacAxis, 
                               const mjtNum point[3], const mjtNum axis[3], int body);

    # Get id of object with specified name, return -1 if not found; type is mjtObj.
    int mj_name2id(const mjModel* m, int type, const char* name);

    # Get name of object with specified id, return 0 if invalid type or id; type is mjtObj.
    const char* mj_id2name(const mjModel* m, int type, int id);

    # Convert sparse inertia matrix M into full (i.e. dense) matrix.
    void mj_fullM(const mjModel* m, mjtNum* dst, const mjtNum* M);

    # Multiply vector by inertia matrix.
    void mj_mulM(const mjModel* m, const mjData* d, mjtNum* res, const mjtNum* vec);

    # Add inertia matrix to destination matrix.
    # Destination can be sparse uncompressed, or dense when all int* are NULL
    void mj_addM(const mjModel* m, mjData* d, mjtNum* dst, 
                       int* rownnz, int* rowadr, int* colind);

    # Apply cartesian force and torque (outside xfrc_applied mechanism).
    void mj_applyFT(const mjModel* m, mjData* d, 
                          const mjtNum* force, const mjtNum* torque, 
                          const mjtNum* point, int body, mjtNum* qfrc_target);

    # Compute object 6D velocity in object-centered frame, world/local orientation.
    void mj_objectVelocity(const mjModel* m, const mjData* d, 
                                 int objtype, int objid, mjtNum* res, int flg_local);

    # Compute object 6D acceleration in object-centered frame, world/local orientation.
    void mj_objectAcceleration(const mjModel* m, const mjData* d, 
                                     int objtype, int objid, mjtNum* res, int flg_local);

    # Compute velocity by finite-differencing two positions.
    void mj_differentiatePos(const mjModel* m, mjtNum* qvel, mjtNum dt,
                                   const mjtNum* qpos1, const mjtNum* qpos2);

    # Extract 6D force:torque for one contact, in contact frame.
    void mj_contactForce(const mjModel* m, const mjData* d, int id, mjtNum* result);

    # Integrate position with given velocity.
    void mj_integratePos(const mjModel* m, mjtNum* qpos, const mjtNum* qvel, mjtNum dt);

    # Normalize all quaterions in qpos-type vector.
    void mj_normalizeQuat(const mjModel* m, mjtNum* qpos);

    # Map from body local to global Cartesian coordinates.
    void mj_local2Global(mjData* d, mjtNum* xpos, mjtNum* xmat, 
                               const mjtNum* pos, const mjtNum* quat, int body);

    # Sum all body masses.
    mjtNum mj_getTotalmass(const mjModel* m);

    # Scale body masses and inertias to achieve specified total mass.
    void mj_setTotalmass(mjModel* m, mjtNum newmass);

    # Return version number: 1.0.2 is encoded as 102.
    int mj_version();


    #--------------------- Ray collisions -------------------------------------------------

    # Intersect ray (pnt+x*vec, x>=0) with visible geoms, except geoms in bodyexclude.
    # Return geomid and distance (x) to nearest surface, or -1 if no intersection.
    # geomgroup, flg_static are as in mjvOption; geomgroup==NULL skips group exclusion.
    mjtNum mj_ray(const mjModel* m, const mjData* d, const mjtNum* pnt, const mjtNum* vec,
                        const mjtByte* geomgroup, mjtByte flg_static, int bodyexclude, 
                        int* geomid);

    # Interect ray with hfield, return nearest distance or -1 if no intersection.
    mjtNum mj_rayHfield(const mjModel* m, const mjData* d, int geomid,
                              const mjtNum* pnt, const mjtNum* vec);

    # Interect ray with mesh, return nearest distance or -1 if no intersection.
    mjtNum mj_rayMesh(const mjModel* m, const mjData* d, int geomid,
                            const mjtNum* pnt, const mjtNum* vec);

    # Interect ray with pure geom, return nearest distance or -1 if no intersection.
    mjtNum mju_rayGeom(const mjtNum* pos, const mjtNum* mat, const mjtNum* size, 
                             const mjtNum* pnt, const mjtNum* vec, int geomtype);


    #--------------------- Interaction ----------------------------------------------------

    # Set default camera.
    void mjv_defaultCamera(mjvCamera* cam);

    # Set default perturbation.
    void mjv_defaultPerturb(mjvPerturb* pert);

    # Transform pose from room to model space.
    void mjv_room2model(mjtNum* modelpos, mjtNum* modelquat, const mjtNum* roompos, 
                              const mjtNum* roomquat, const mjvScene* scn);

    # Transform pose from model to room space.
    void mjv_model2room(mjtNum* roompos, mjtNum* roomquat, const mjtNum* modelpos, 
                              const mjtNum* modelquat, const mjvScene* scn);

    # Get camera info in model space; average left and right OpenGL cameras.
    void mjv_cameraInModel(mjtNum* headpos, mjtNum* forward, mjtNum* up,
                                 const mjvScene* scn);

    # Get camera info in room space; average left and right OpenGL cameras.
    void mjv_cameraInRoom(mjtNum* headpos, mjtNum* forward, mjtNum* up,
                                const mjvScene* scn);

    # Get frustum height at unit distance from camera; average left and right OpenGL cameras.
    mjtNum mjv_frustumHeight(const mjvScene* scn);

    # Rotate 3D vec in horizontal plane by angle between (0,1) and (forward_x,forward_y).
    void mjv_alignToCamera(mjtNum* res, const mjtNum* vec, const mjtNum* forward);

    # Move camera with mouse; action is mjtMouse.
    void mjv_moveCamera(const mjModel* m, int action, mjtNum reldx, mjtNum reldy, 
                              const mjvScene* scn, mjvCamera* cam);

    # Move perturb object with mouse; action is mjtMouse.
    void mjv_movePerturb(const mjModel* m, const mjData* d, int action, mjtNum reldx, 
                               mjtNum reldy, const mjvScene* scn, mjvPerturb* pert);

    # Move model with mouse; action is mjtMouse.
    void mjv_moveModel(const mjModel* m, int action, mjtNum reldx, mjtNum reldy,
                             const mjtNum* roomup, mjvScene* scn);

    # Copy perturb pos,quat from selected body; set scale for perturbation.
    void mjv_initPerturb(const mjModel* m, const mjData* d, 
                               const mjvScene* scn, mjvPerturb* pert);

    # Set perturb pos,quat in d->mocap when selected body is mocap, and in d->qpos otherwise.
    # Write d->qpos only if flg_paused and subtree root for selected body has free joint.
    void mjv_applyPerturbPose(const mjModel* m, mjData* d, const mjvPerturb* pert, 
                                    int flg_paused);

    # Set perturb force,torque in d->xfrc_applied, if selected body is dynamic.
    void mjv_applyPerturbForce(const mjModel* m, mjData* d, const mjvPerturb* pert);

    # Return the average of two OpenGL cameras.
    mjvGLCamera mjv_averageCamera(const mjvGLCamera* cam1, const mjvGLCamera* cam2);

    # Select model geom with mouse, return -1 if none selected. selpnt is the 3D point.
    int mjv_select(const mjModel* m, const mjData* d, const mjvOption* vopt,
                         mjtNum aspectratio, mjtNum relx, mjtNum rely, 
                         const mjvScene* scn, mjtNum* selpnt);


    #--------------------- Visualization --------------------------------------------------

    # Set default visualization options.
    void mjv_defaultOption(mjvOption* opt);

    # Set default figure.
    void mjv_defaultFigure(mjvFigure* fig);

    # Initialize given geom fields when not NULL, set the rest to their default values.
    void mjv_initGeom(mjvGeom* geom, int type, const mjtNum* size,
                            const mjtNum* pos, const mjtNum* mat, const float* rgba);

    # Set (type, size, pos, mat) for connector-type geom between given points.
    # Assume that mjv_initGeom was already called to set all other properties.
    void mjv_makeConnector(mjvGeom* geom, int type, mjtNum width, 
                                 mjtNum a0, mjtNum a1, mjtNum a2, 
                                 mjtNum b0, mjtNum b1, mjtNum b2);

    # Allocate and init abstract scene.
    void mjv_makeScene(mjvScene* scn, int maxgeom);

    # Free abstract scene.
    void mjv_freeScene(mjvScene* scn);

    # Update entire scene given model state.
    void mjv_updateScene(const mjModel* m, mjData* d, const mjvOption* opt, 
                               const mjvPerturb* pert, mjvCamera* cam, int catmask, mjvScene* scn);

    # Add geoms from selected categories to existing scene.
    void mjv_addGeoms(const mjModel* m, mjData* d, const mjvOption* opt, 
                            const mjvPerturb* pert, int catmask, mjvScene* scn);

    # Update camera only.
    void mjv_updateCamera(const mjModel* m, mjData* d, mjvCamera* cam, mjvScene* scn);


    #--------------------- OpenGL rendering -----------------------------------------------

    # Set default mjrContext.
    void mjr_defaultContext(mjrContext* con);

    # Allocate resources in custom OpenGL context; fontscale is mjtFontScale.
    void mjr_makeContext(const mjModel* m, mjrContext* con, int fontscale);

    # Free resources in custom OpenGL context, set to default.
    void mjr_freeContext(mjrContext* con);

    # Upload texture to GPU, overwriting previous upload if any.
    void mjr_uploadTexture(const mjModel* m, const mjrContext* con, int texid);

    # Upload mesh to GPU, overwriting previous upload if any.
    void mjr_uploadMesh(const mjModel* m, const mjrContext* con, int meshid);

    # Upload height field to GPU, overwriting previous upload if any.
    void mjr_uploadHField(const mjModel* m, const mjrContext* con, int hfieldid);

    # Set OpenGL framebuffer for rendering: mjFB_WINDOW or mjFB_OFFSCREEN.
    # If only one buffer is available, set that buffer and ignore framebuffer argument.
    void mjr_setBuffer(int framebuffer, mjrContext* con);

    # Read pixels from current OpenGL framebuffer to client buffer.
    # Viewport is in OpenGL framebuffer; client buffer starts at (0,0).
    void mjr_readPixels(unsigned char* rgb, float* depth,
                              mjrRect viewport, const mjrContext* con);

    # Draw pixels from client buffer to current OpenGL framebuffer.
    # Viewport is in OpenGL framebuffer; client buffer starts at (0,0).
    void mjr_drawPixels(const unsigned char* rgb, const float* depth,
                              mjrRect viewport, const mjrContext* con);

    # Blit from src viewpoint in current framebuffer to dst viewport in other framebuffer.
    # If src, dst have different size and flg_depth==0, color is interpolated with GL_LINEAR.
    void mjr_blitBuffer(mjrRect src, mjrRect dst, 
                              int flg_color, int flg_depth, const mjrContext* con);

    # Draw text at (x,y) in relative coordinates; font is mjtFont.
    void mjr_text(int font, const char* txt, const mjrContext* con,
                        float x, float y, float r, float g, float b);

    # Draw text overlay; font is mjtFont; gridpos is mjtGridPos.
    void mjr_overlay(int font, int gridpos, mjrRect viewport,
                           const char* overlay, const char* overlay2, const mjrContext* con);

    # Get maximum viewport for active buffer.
    mjrRect mjr_maxViewport(const mjrContext* con);

    # Draw rectangle.
    void mjr_rectangle(mjrRect viewport, float r, float g, float b, float a);

    # Draw 2D figure.
    void mjr_figure(mjrRect viewport, const mjvFigure* fig, const mjrContext* con);

    # Render 3D scene.
    void mjr_render(mjrRect viewport, mjvScene* scn, const mjrContext* con);

    # Call glFinish.
    void mjr_finish();

    # Call glGetError and return result.
    int mjr_getError();


    #--------------------- Error and memory -----------------------------------------------

    # Main error function; does not return to caller.
    void mju_error(const char* msg);

    # Error function with int argument; msg is a printf format string.
    void mju_error_i(const char* msg, int i);

    # Error function with string argument.
    void mju_error_s(const char* msg, const char* text);

    # Main warning function; returns to caller.
    void mju_warning(const char* msg);

    # Warning function with int argument.
    void mju_warning_i(const char* msg, int i);

    # Warning function with string argument.
    void mju_warning_s(const char* msg, const char* text);

    # Clear user error and memory handlers.
    void mju_clearHandlers();

    # Allocate memory; byte-align on 8; pad size to multiple of 8.
    void* mju_malloc(size_t size);

    # Free memory, using free() by default.
    void mju_free(void* ptr);

    # High-level warning function: count warnings in mjData, print only the first.
    void mj_warning(mjData* d, int warning, int info);

    # Write [datetime, type: message] to MUJOCO_LOG.TXT.
    void mju_writeLog(const char* type, const char* msg);


    #--------------------- Standard math --------------------------------------------------

#define mjMAX(a,b) (((a) > (b)) ? (a) : (b))
#define mjMIN(a,b) (((a) < (b)) ? (a) : (b))

#ifdef mjUSEDOUBLE
        #define mju_sqrt    sqrt
        #define mju_exp     exp
        #define mju_sin     sin
        #define mju_cos     cos
        #define mju_tan     tan
        #define mju_asin    asin
        #define mju_acos    acos
        #define mju_atan2   atan2
        #define mju_tanh    tanh
        #define mju_pow     pow
        #define mju_abs     fabs
        #define mju_log     log
        #define mju_log10   log10
        #define mju_floor   floor
        #define mju_ceil    ceil

#else
        #define mju_sqrt    sqrtf
        #define mju_exp     expf
        #define mju_sin     sinf
        #define mju_cos     cosf
        #define mju_tan     tanf
        #define mju_asin    asinf
        #define mju_acos    acosf
        #define mju_atan2   atan2f
        #define mju_tanh    tanhf
        #define mju_pow     powf
        #define mju_abs     fabsf
        #define mju_log     logf
        #define mju_log10   log10f
        #define mju_floor   floorf
        #define mju_ceil    ceilf
#endif


    #----------------------------- Vector math --------------------------------------------

    # Set res = 0.
    void mju_zero3(mjtNum res[3]);

    # Set res = vec.
    void mju_copy3(mjtNum res[3], const mjtNum data[3]);

    # Set res = vec*scl.
    void mju_scl3(mjtNum res[3], const mjtNum vec[3], mjtNum scl);

    # Set res = vec1 + vec2.
    void mju_add3(mjtNum res[3], const mjtNum vec1[3], const mjtNum vec2[3]);

    # Set res = vec1 - vec2.
    void mju_sub3(mjtNum res[3], const mjtNum vec1[3], const mjtNum vec2[3]);

    # Set res = res + vec.
    void mju_addTo3(mjtNum res[3], const mjtNum vec[3]);

    # Set res = res - vec.
    void mju_subFrom3(mjtNum res[3], const mjtNum vec[3]);

    # Set res = res + vec*scl.
    void mju_addToScl3(mjtNum res[3], const mjtNum vec[3], mjtNum scl);

    # Set res = vec1 + vec2*scl.
    void mju_addScl3(mjtNum res[3], const mjtNum vec1[3], const mjtNum vec2[3], mjtNum scl);

    # Normalize vector, return length before normalization.
    mjtNum mju_normalize3(mjtNum res[3]);

    # Return vector length (without normalizing the vector).
    mjtNum mju_norm3(const mjtNum vec[3]);

    # Return dot-product of vec1 and vec2.
    mjtNum mju_dot3(const mjtNum vec1[3], const mjtNum vec2[3]);

    # Return Cartesian distance between 3D vectors pos1 and pos2.
    mjtNum mju_dist3(const mjtNum pos1[3], const mjtNum pos2[3]);

    # Multiply vector by 3D rotation matrix: res = mat * vec.
    void mju_rotVecMat(mjtNum res[3], const mjtNum vec[3], const mjtNum mat[9]);

    # Multiply vector by transposed 3D rotation matrix: res = mat' * vec.
    void mju_rotVecMatT(mjtNum res[3], const mjtNum vec[3], const mjtNum mat[9]);

    # Compute cross-product: res = cross(a, b).
    void mju_cross(mjtNum res[3], const mjtNum a[3], const mjtNum b[3]);

    # Set res = 0.
    void mju_zero4(mjtNum res[4]);

    # Set res = (1,0,0,0).
    void mju_unit4(mjtNum res[4]);

    # Set res = vec.
    void mju_copy4(mjtNum res[4], const mjtNum data[4]);

    # Normalize vector, return length before normalization.
    mjtNum mju_normalize4(mjtNum res[4]);

    # Set res = 0.
    void mju_zero(mjtNum* res, int n);

    # Set res = vec.
    void mju_copy(mjtNum* res, const mjtNum* data, int n);

    # Set res = vec*scl.
    void mju_scl(mjtNum* res, const mjtNum* vec, mjtNum scl, int n);

    # Set res = vec1 + vec2.
    void mju_add(mjtNum* res, const mjtNum* vec1, const mjtNum* vec2, int n);

    # Set res = vec1 - vec2.
    void mju_sub(mjtNum* res, const mjtNum* vec1, const mjtNum* vec2, int n);

    # Set res = res + vec.
    void mju_addTo(mjtNum* res, const mjtNum* vec, int n);

    # Set res = res - vec.
    void mju_subFrom(mjtNum* res, const mjtNum* vec, int n);

    # Set res = res + vec*scl.
    void mju_addToScl(mjtNum* res, const mjtNum* vec, mjtNum scl, int n);

    # Set res = vec1 + vec2*scl.
    void mju_addScl(mjtNum* res, const mjtNum* vec1, const mjtNum* vec2, mjtNum scl, int n);

    # Normalize vector, return length before normalization.
    mjtNum mju_normalize(mjtNum* res, int n);

    # Return vector length (without normalizing vector).
    mjtNum mju_norm(const mjtNum* res, int n);

    # Return dot-product of vec1 and vec2.
    mjtNum mju_dot(const mjtNum* vec1, const mjtNum* vec2, const int n);

    # Multiply matrix and vector: res = mat * vec.
    void mju_mulMatVec(mjtNum* res, const mjtNum* mat, const mjtNum* vec,
                             int nr, int nc);

    # Multiply transposed matrix and vector: res = mat' * vec.
    void mju_mulMatTVec(mjtNum* res, const mjtNum* mat, const mjtNum* vec,
                              int nr, int nc);

    # Transpose matrix: res = mat'.
    void mju_transpose(mjtNum* res, const mjtNum* mat, int nr, int nc);

    # Multiply matrices: res = mat1 * mat2.
    void mju_mulMatMat(mjtNum* res, const mjtNum* mat1, const mjtNum* mat2,
                             int r1, int c1, int c2);

    # Multiply matrices, second argument transposed: res = mat1 * mat2'.
    void mju_mulMatMatT(mjtNum* res, const mjtNum* mat1, const mjtNum* mat2,
                              int r1, int c1, int r2);

    # Multiply matrices, first argument transposed: res = mat1' * mat2.
    void mju_mulMatTMat(mjtNum* res, const mjtNum* mat1, const mjtNum* mat2,
                              int r1, int c1, int c2);

    # Set res = mat' * diag * mat if diag is not NULL, and res = mat' * mat otherwise.
    void mju_sqrMatTD(mjtNum* res, const mjtNum* mat, const mjtNum* diag, int nr, int nc);

    # Coordinate transform of 6D motion or force vector in rotation:translation format.
    # rotnew2old is 3-by-3, NULL means no rotation; flg_force specifies force or motion type.
    void mju_transformSpatial(mjtNum res[6], const mjtNum vec[6], int flg_force,
                                    const mjtNum newpos[3], const mjtNum oldpos[3], 
                                    const mjtNum rotnew2old[9]);


    #--------------------- Sparse math ----------------------------------------------------

    # Return dot-product of vec1 and vec2, where vec1 is sparse.
    mjtNum mju_dotSparse(const mjtNum* vec1, const mjtNum* vec2,
                               const int nnz1, const int* ind1);

    # Return dot-product of vec1 and vec2, where both vectors are sparse.
    mjtNum mju_dotSparse2(const mjtNum* vec1, const mjtNum* vec2,
                                const int nnz1, const int* ind1,
                                const int nnz2, const int* ind2);

    # Convert matrix from dense to sparse format.
    void mju_dense2sparse(mjtNum* res, const mjtNum* mat, int nr, int nc,
                                int* rownnz, int* rowadr, int* colind);

    # Convert matrix from sparse to dense format.
    void mju_sparse2dense(mjtNum* res, const mjtNum* mat, int nr, int nc,
                                const int* rownnz, const int* rowadr, const int* colind);

    # Multiply sparse matrix and dense vector:  res = mat * vec.
    void mju_mulMatVecSparse(mjtNum* res, const mjtNum* mat, const mjtNum* vec, int nr, 
                                   const int* rownnz, const int* rowadr, const int* colind);

    # Compress layout of sparse matrix.
    void mju_compressSparse(mjtNum* mat, int nr, int nc, 
                                  int* rownnz, int* rowadr, int* colind);

    # Set dst = a*dst + b*src, return nnz of result, modify dst sparsity pattern as needed.
    # Both vectors are sparse. The required scratch space is 2*n.
    int mju_combineSparse(mjtNum* dst, const mjtNum* src, int n, mjtNum a, mjtNum b,
                                int dst_nnz, int src_nnz, int* dst_ind, const int* src_ind, 
                                mjtNum* scratch, int nscratch);

    # Set res = matT * diag * mat if diag is not NULL, and res = matT * mat otherwise.
    # The required scratch space is 3*nc. The result has uncompressed layout.
    void mju_sqrMatTDSparse(mjtNum* res, const mjtNum* mat, const mjtNum* matT, 
                                  const mjtNum* diag, int nr, int nc, 
                                  int* res_rownnz, int* res_rowadr, int* res_colind,
                                  const int* rownnz, const int* rowadr, const int* colind,
                                  const int* rownnzT, const int* rowadrT, const int* colindT,
                                  mjtNum* scratch, int nscratch);

    # Transpose sparse matrix.
    void mju_transposeSparse(mjtNum* res, const mjtNum* mat, int nr, int nc,
                                  int* res_rownnz, int* res_rowadr, int* res_colind,
                                  const int* rownnz, const int* rowadr, const int* colind);


    #--------------------- Quaternions ----------------------------------------------------

    # Rotate vector by quaternion.
    void mju_rotVecQuat(mjtNum res[3], const mjtNum vec[3], const mjtNum quat[4]);

    # Negate quaternion.
    void mju_negQuat(mjtNum res[4], const mjtNum quat[4]);

    # Muiltiply quaternions.
    void mju_mulQuat(mjtNum res[4], const mjtNum quat1[4], const mjtNum quat2[4]);

    # Muiltiply quaternion and axis.
    void mju_mulQuatAxis(mjtNum res[4], const mjtNum quat[4], const mjtNum axis[3]);

    # Convert axisAngle to quaternion.
    void mju_axisAngle2Quat(mjtNum res[4], const mjtNum axis[3], mjtNum angle);

    # Convert quaternion (corresponding to orientation difference) to 3D velocity.
    void mju_quat2Vel(mjtNum res[3], const mjtNum quat[4], mjtNum dt);

    # Convert quaternion to 3D rotation matrix.
    void mju_quat2Mat(mjtNum res[9], const mjtNum quat[4]);

    # Convert 3D rotation matrix to quaterion.
    void mju_mat2Quat(mjtNum quat[4], const mjtNum mat[9]);

    # Compute time-derivative of quaternion, given 3D rotational velocity.
    void mju_derivQuat(mjtNum res[4], const mjtNum quat[4], const mjtNum vel[3]);

    # Integrate quaterion given 3D angular velocity.
    void mju_quatIntegrate(mjtNum quat[4], const mjtNum vel[3], mjtNum scale);

    # Construct quaternion performing rotation from z-axis to given vector.
    void mju_quatZ2Vec(mjtNum quat[4], const mjtNum vec[3]);


    #--------------------- Poses ----------------------------------------------------------

    # Multiply two poses.
    void mju_mulPose(mjtNum posres[3], mjtNum quatres[4], 
                           const mjtNum pos1[3], const mjtNum quat1[4], 
                           const mjtNum pos2[3], const mjtNum quat2[4]);

    # Negate pose.
    void mju_negPose(mjtNum posres[3], mjtNum quatres[4],
                           const mjtNum pos[3], const mjtNum quat[4]);

    # Transform vector by pose.
    void mju_trnVecPose(mjtNum res[3], const mjtNum pos[3], const mjtNum quat[4],
                              const mjtNum vec[3]);


    #--------------------- Decompositions --------------------------------------------------

    # Cholesky decomposition: mat = L*L'; return rank.
    int mju_cholFactor(mjtNum* mat, int n);

    # Solve mat * res = vec, where mat is Cholesky-factorized
    void mju_cholSolve(mjtNum* res, const mjtNum* mat, const mjtNum* vec, int n);

    # Cholesky rank-one update: L*L' +/- x*x'; return rank.
    int mju_cholUpdate(mjtNum* mat, mjtNum* x, int n, int flg_plus);

    # Sparse reverse-order Cholesky decomposition: mat = L'*L; return 'rank'.
    # mat must have uncompressed layout; rownnz is modified to end at diagonal.
    # The required scratch space is 2*n.
    int mju_cholFactorSparse(mjtNum* mat, int n, 
                                   int* rownnz, int* rowadr, int* colind,
                                   mjtNum* scratch, int nscratch);

    # Solve mat * res = vec, where mat is sparse reverse-order Cholesky factorized.
    void mju_cholSolveSparse(mjtNum* res, const mjtNum* mat, const mjtNum* vec, int n,
                                   const int* rownnz, const int* rowadr, const int* colind);

    # Sparse reverse-order Cholesky rank-one update: L'*L +/- x*x'; return rank.
    # The vector x is sparse; changes in sparsity pattern of mat are not allowed.
    # The required scratch space is 2*n.
    int mju_cholUpdateSparse(mjtNum* mat, mjtNum* x, int n, int flg_plus,
                                   int* rownnz, int* rowadr, int* colind, int x_nnz, int* x_ind, 
                                   mjtNum* scratch, int nscratch);

    # Eigenvalue decomposition of symmetric 3x3 matrix.
    int mju_eig3(mjtNum* eigval, mjtNum* eigvec, mjtNum* quat, const mjtNum* mat);


    #--------------------- Miscellaneous --------------------------------------------------

    # Muscle model; not yet implemented.
    mjtNum mju_muscleFVL(mjtNum len, mjtNum vel, mjtNum lmin, mjtNum lmax, mjtNum* prm);

    # Passive muscle force; not yet implemented.
    mjtNum mju_musclePassive(mjtNum len, mjtNum lmin, mjtNum lmax, mjtNum* prm);

    # Pneumatic cylinder dynamics; not yet implemented.
    mjtNum mju_pneumatic(mjtNum len, mjtNum len0, mjtNum vel, mjtNum* prm,
                               mjtNum act, mjtNum ctrl, mjtNum timestep, mjtNum* jac);

    # Convert contact force to pyramid representation.
    void mju_encodePyramid(mjtNum* pyramid, const mjtNum* force,
                                 const mjtNum* mu, int dim);

    # Convert pyramid representation to contact force.
    void mju_decodePyramid(mjtNum* force, const mjtNum* pyramid,
                                 const mjtNum* mu, int dim);

    # Integrate spring-damper analytically, return pos(dt).
    mjtNum mju_springDamper(mjtNum pos0, mjtNum vel0, mjtNum Kp, mjtNum Kv, mjtNum dt);

    # Return min(a,b) with single evaluation of a and b.
    mjtNum mju_min(mjtNum a, mjtNum b);

    # Return max(a,b) with single evaluation of a and b.
    mjtNum mju_max(mjtNum a, mjtNum b);

    # Return sign of x: +1, -1 or 0.
    mjtNum mju_sign(mjtNum x);

    # Round x to nearest integer.
    int mju_round(mjtNum x);

    # Convert type id (mjtObj) to type name.
    const char* mju_type2Str(int type);

    # Convert type name to type id (mjtObj).
    int mju_str2Type(const char* str);

    # Construct a warning message given the warning type and info.
    const char* mju_warningText(int warning, int info);

    # Return 1 if nan or abs(x)>mjMAXVAL, 0 otherwise. Used by check functions.
    int mju_isBad(mjtNum x);

    # Return 1 if all elements are 0.
    int mju_isZero(mjtNum* vec, int n);

    # Standard normal random number generator (optional second number).
    mjtNum mju_standardNormal(mjtNum* num2);

    # Convert from float to mjtNum.
    void mju_f2n(mjtNum* res, const float* vec, int n);

    # Convert from mjtNum to float.
    void mju_n2f(float* res, const mjtNum* vec, int n);

    # Convert from double to mjtNum.
    void mju_d2n(mjtNum* res, const double* vec, int n);

    # Convert from mjtNum to double.
    void mju_n2d(double* res, const mjtNum* vec, int n);

    # Insertion sort, resulting list is in increasing order.
    void mju_insertionSort(mjtNum* list, int n);

    # Generate Halton sequence.
    mjtNum mju_Halton(int index, int base);




include "mjmodel.pxd"


cdef extern from "mjdata.h" nogil:

    #---------------------------- primitive types (mjt) ------------------------------------
        
    ctypedef enum mjtWarning:            # warning types
        mjWARN_INERTIA      = 0,        # (near) singular inertia matrix
        mjWARN_CONTACTFULL,             # too many contacts in contact list
        mjWARN_CNSTRFULL,               # too many constraints
        mjWARN_VGEOMFULL,               # too many visual geoms
        mjWARN_BADQPOS,                 # bad number in qpos
        mjWARN_BADQVEL,                 # bad number in qvel
        mjWARN_BADQACC,                 # bad number in qacc
        mjWARN_BADCTRL,                 # bad number in ctrl

    enum: mjNWARNING                      # number of warnings


    ctypedef enum mjtTimer:
        # main api
        mjTIMER_STEP        = 0,        # step
        mjTIMER_FORWARD,                # forward
        mjTIMER_INVERSE,                # inverse

        # breakdown of step/forward
        mjTIMER_POSITION,               # fwdPosition
        mjTIMER_VELOCITY,               # fwdVelocity
        mjTIMER_ACTUATION,              # fwdActuation
        mjTIMER_ACCELERATION,           # fwdAcceleration
        mjTIMER_CONSTRAINT,             # fwdConstraint

        # breakdown of fwdPosition
        mjTIMER_POS_KINEMATICS,         # kinematics, com, tendon, transmission
        mjTIMER_POS_INERTIA,            # inertia computations
        mjTIMER_POS_COLLISION,          # collision detection
        mjTIMER_POS_MAKE,               # make constraints
        mjTIMER_POS_PROJECT,            # project constraints

    enum: mjNTIMER                        # number of timers


    #------------------------------ mjContact ----------------------------------------------

    ctypedef struct mjContact:                   # result of collision detection functions
        # contact parameters set by geom-specific collision detector
        mjtNum dist                     # distance between nearest points; neg: penetration
        mjtNum pos[3]                   # position of contact point: midpoint between geoms
        mjtNum frame[9]                 # normal is in [0-2]

        # contact parameters set by mj_collideGeoms
        mjtNum includemargin            # include if dist<includemargin=margin-gap
        mjtNum friction[5]              # tangent1, 2, spin, roll1, 2
        mjtNum solref[mjNREF]           # constraint solver reference
        mjtNum solimp[mjNIMP]           # constraint solver impedance

        # storage used internally by constraint solver
        mjtNum mu                       # friction of regularized cone
        mjtNum H[36]                    # cone Hessian, set by mj_updateConstraint

        # contact descriptors set by mj_collideGeoms
        int dim                         # contact space dimensionality: 1, 3, 4 or 6
        int geom1                       # id of geom 1
        int geom2                       # id of geom 2

        # flag set by mj_fuseContact or mj_instantianteEquality
        int exclude                     # 0: include, 1: in gap, 2: fused, 3: equality

        # address computed by mj_instantiateContact
        int efc_address                 # address in efc; -1: not included, -2-i: distance constraint i ???

    #------------------------------ diagnostics --------------------------------------------

    ctypedef struct mjWarningStat:      # warning statistics
        int lastinfo                    # info from last warning
        int number                      # how many times was warning raised


    ctypedef struct mjTimerStat:        # timer statistics
        mjtNum duration                 # cumulative duration
        int number                      # how many times was timer called


    ctypedef struct mjSolverStat:       # per-iteration solver statistics
        mjtNum improvement              # cost reduction, scaled by 1/trace(M(qpos0))
        mjtNum gradient                 # gradient norm (primal only, scaled)
        mjtNum lineslope                # slope in linesearch
        int nactive                     # number of active constraints
        int nchange                     # number of constraint state changes
        int neval                       # number of cost evaluations in line search
        int nupdate                     # number of Cholesky updates in line search



    #---------------------------------- mjData ---------------------------------------------
    ctypedef struct mjData:
        # constant sizes
        int nstack                      # number of mjtNums that can fit in stack
        int nbuffer                     # size of main buffer in bytes

        # stack pointer
        int pstack                      # first available mjtNum address in stack

        # memory utilization stats
        int maxuse_stack                # maximum stack allocation
        int maxuse_con                  # maximum number of contacts
        int maxuse_efc                  # maximum number of scalar constraints

        # diagnostics
        mjWarningStat warning[mjNWARNING] # warning statistics
        mjTimerStat timer[mjNTIMER]       # timer statistics
        mjSolverStat solver[mjNSOLVER]    # solver statistics per iteration
        int solver_iter                   # number of solver iterations
        int solver_nnz                  # number of non-zeros in Hessian or efc_AR
        mjtNum solver_fwdinv[2]         # forward-inverse comparison: qfrc, efc

        # variable sizes
        int ne                          # number of equality constraints
        int nf                          # number of friction constraints
        int nefc                        # number of constraints
        int ncon                        # number of detected contacts

        # global properties
        mjtNum time                     # simulation time
        mjtNum energy[2]                # potential, kinetic energy

        #-------------------------------- end of info header

        # buffers
        void*     buffer                # main buffer; all pointers point in it    (nbuffer bytes)
        mjtNum*   stack                 # stack buffer                             (nstack mjtNums)

        #-------------------------------- main inputs and outputs of the computation

        # state
        mjtNum*   qpos                  # position                                 (nq x 1)
        mjtNum*   qvel                  # velocity                                 (nv x 1)
        mjtNum*   act                   # actuator activation                      (na x 1)

        # control
        mjtNum*   ctrl                  # control                                  (nu x 1)
        mjtNum*   qfrc_applied          # applied generalized force                (nv x 1)
        mjtNum*   xfrc_applied          # applied Cartesian force/torque           (nbody x 6)

        # dynamics
        mjtNum*   qacc                  # acceleration                             (nv x 1)
        mjtNum*   act_dot               # time-derivative of actuator activation   (na x 1)

        # mocap data
        mjtNum*  mocap_pos              # positions of mocap bodies                (nmocap x 3)
        mjtNum*  mocap_quat             # orientations of mocap bodies             (nmocap x 4)

        # user data
        mjtNum*  userdata               # user data, not touched by engine         (nuserdata x 1)

        # sensors
        mjtNum*  sensordata             # sensor data array                        (nsensordata x 1)

        #-------------------------------- POSITION dependent

        # computed by mj_fwdPosition/mj_kinematics
        mjtNum*   xpos                  # Cartesian position of body frame         (nbody x 3)
        mjtNum*   xquat                 # Cartesian orientation of body frame      (nbody x 4)
        mjtNum*   xmat                  # Cartesian orientation of body frame      (nbody x 9)
        mjtNum*   xipos                 # Cartesian position of body com           (nbody x 3)
        mjtNum*   ximat                 # Cartesian orientation of body inertia    (nbody x 9)
        mjtNum*   xanchor               # Cartesian position of joint anchor       (njnt x 3)
        mjtNum*   xaxis                 # Cartesian joint axis                     (njnt x 3)
        mjtNum*   geom_xpos             # Cartesian geom position                  (ngeom x 3)
        mjtNum*   geom_xmat             # Cartesian geom orientation               (ngeom x 9)
        mjtNum*   site_xpos             # Cartesian site position                  (nsite x 3)
        mjtNum*   site_xmat             # Cartesian site orientation               (nsite x 9)
        mjtNum*   cam_xpos              # Cartesian camera position                (ncam x 3)
        mjtNum*   cam_xmat              # Cartesian camera orientation             (ncam x 9)
        mjtNum*   light_xpos            # Cartesian light position                 (nlight x 3)
        mjtNum*   light_xdir            # Cartesian light direction                (nlight x 3)

        # computed by mj_fwdPosition/mj_comPos
        mjtNum*   subtree_com           # center of mass of each subtree           (nbody x 3)
        mjtNum*   cdof                  # com-based motion axis of each dof        (nv x 6)
        mjtNum*   cinert                # com-based body inertia and mass          (nbody x 10)

        # computed by mj_fwdPosition/mj_tendon
        int*      ten_wrapadr           # start address of tendon's path           (ntendon x 1)
        int*      ten_wrapnum           # number of wrap points in path            (ntendon x 1)
        mjtNum*   ten_length            # tendon lengths                           (ntendon x 1)
        mjtNum*   ten_moment            # tendon moment arms                       (ntendon x nv)
        int*      wrap_obj              # geom id; -1: site; -2: pulley            (nwrap*2 x 1)
        mjtNum*   wrap_xpos             # Cartesian 3D points in all path          (nwrap*2 x 3)

        # computed by mj_fwdPosition/mj_transmission
        mjtNum*   actuator_length       # actuator lengths                         (nu x 1)
        mjtNum*   actuator_moment       # actuator moment arms                     (nu x nv)

        # computed by mj_fwdPosition/mj_crb
        mjtNum*   crb                   # com-based composite inertia and mass     (nbody x 10)
        mjtNum*   qM                    # total inertia                            (nM x 1)

        # computed by mj_fwdPosition/mj_factorM
        mjtNum*   qLD                   # L'*D*L factorization of M                (nM x 1)
        mjtNum*   qLDiagInv             # 1/diag(D)                                (nv x 1)
        mjtNum*   qLDiagSqrtInv         # 1/sqrt(diag(D))                          (nv x 1)

        # computed by mj_fwdPosition/mj_collision
        mjContact* contact              # list of all detected contacts            (nconmax x 1)

        # computed by mj_fwdPosition/mj_makeConstraint
        int*      efc_type              # constraint type (mjtConstraint)          (njmax x 1)
        int*      efc_id                # id of object of specified type           (njmax x 1)
        int*      efc_J_rownnz          # number of non-zeros in Jacobian row      (njmax x 1)
        int*      efc_J_rowadr          # row start address in colind array        (njmax x 1)
        int*      efc_J_colind          # column indices in sparse Jacobian        (njmax x nv)
        int*      efc_JT_rownnz         # number of non-zeros in Jacobian row  T   (nv x 1)
        int*      efc_JT_rowadr         # row start address in colind array    T   (nv x 1)
        int*      efc_JT_colind         # column indices in sparse Jacobian    T   (nv x njmax)
        mjtNum*   efc_solref            # constraint solver reference              (njmax x mjNREF)
        mjtNum*   efc_solimp            # constraint solver impedance              (njmax x mjNIMP)
        mjtNum*   efc_margin            # inclusion margin (contact)               (njmax x 1)
        mjtNum*   efc_frictionloss      # frictionloss (friction)                  (njmax x 1)
        mjtNum*   efc_pos               # constraint position (equality, contact)  (njmax x 1)
        mjtNum*   efc_J                 # constraint Jacobian                      (njmax x nv)
        mjtNum*   efc_JT                # sparse constraint Jacobian transposed    (nv x njmax)
        mjtNum*   efc_diagApprox        # approximation to diagonal of A           (njmax x 1)
        mjtNum*   efc_D                 # constraint mass                          (njmax x 1)
        mjtNum*   efc_R                 # inverse constraint mass                  (njmax x 1)

        # computed by mj_fwdPosition/mj_projectConstraint
        int*      efc_AR_rownnz         # number of non-zeros in AR                (njmax x 1)
        int*      efc_AR_rowadr         # row start address in colind array        (njmax x 1)
        int*      efc_AR_colind         # column indices in sparse AR              (njmax x njmax)
        mjtNum*   efc_AR                # J*inv(M)*J' + R                          (njmax x njmax)

        #-------------------------------- POSITION, VELOCITY dependent

        # computed by mj_fwdVelocity
        mjtNum*   ten_velocity          # tendon velocities                        (ntendon x 1)
        mjtNum*   actuator_velocity     # actuator velocities                      (nu x 1)

        # computed by mj_fwdVelocity/mj_comVel
        mjtNum*   cvel                  # com-based velocity [3D rot; 3D tran]     (nbody x 6)
        mjtNum*   cdof_dot              # time-derivative of cdof                  (nv x 6)

        # computed by mj_fwdVelocity/mj_rne (without acceleration)
        mjtNum*   qfrc_bias             # C(qpos,qvel)                             (nv x 1)

        # computed by mj_fwdVelocity/mj_passive
        mjtNum*   qfrc_passive          # passive force                            (nv x 1)

        # computed by mj_fwdVelocity/mj_referenceConstraint
        mjtNum*   efc_vel               # velocity in constraint space: J*qvel     (njmax x 1)
        mjtNum*   efc_aref              # reference pseudo-acceleration            (njmax x 1)

        # computed by mj_sensorVel
        mjtNum*   subtree_linvel        # linear velocity of subtree com           (nbody x 3)
        mjtNum*   subtree_angmom        # angular momentum about subtree com       (nbody x 3)

        #-------------------------------- POSITION, VELOCITY, CONTROL/ACCELERATION dependent

        # computed by mj_fwdActuation
        mjtNum*   actuator_force        # actuator force in actuation space        (nu x 1)
        mjtNum*   qfrc_actuator         # actuator force                           (nv x 1)

        # computed by mj_fwdAcceleration
        mjtNum*   qfrc_unc              # net unconstrained force                  (nv x 1)
        mjtNum*   qacc_unc              # unconstrained acceleration               (nv x 1)

        # computed by mj_fwdConstraint/mj_inverse
        mjtNum*   efc_b                 # linear cost term: J*qacc_unc - aref      (njmax x 1)
        mjtNum*   efc_force             # constraint force in constraint space     (njmax x 1)
        int*      efc_state             # constraint state (mjtConstraintState)    (njmax x 1)
        mjtNum*   qfrc_constraint       # constraint force                         (nv x 1)
        mjtNum*   qacc_warmstart        # acceleration used for warmstart          (nv x 1)

        # computed by mj_inverse
        mjtNum*   qfrc_inverse          # net external force; should equal:        (nv x 1)
                                        #  qfrc_applied + J'*xfrc_applied + qfrc_actuator

        # computed by mj_sensorAcc/mj_rnePostConstraint; rotation:translation format
        mjtNum*   cacc                  # com-based acceleration                   (nbody x 6)
        mjtNum*   cfrc_int              # com-based interaction force with parent  (nbody x 6)
        mjtNum*   cfrc_ext              # com-based external force on body         (nbody x 6)


# #---------------------------------- callback function types ----------------------------
#
# # generic MuJoCo function
# typedef void (*mjfGeneric)(const mjModel* m, mjData* d)
#
# # sensor simulation
# typedef void (*mjfSensor)(const mjModel* m, mjData* d, int stage)
#
# # timer
# typedef long long int (*mjfTime)(void);
#
# # actuator dynamics, gain, bias
# typedef mjtNum (*mjfAct)(const mjModel* m, const mjData* d, int id);
#
# # solver impedance
# typedef mjtNum (*mjfSolImp)(const mjModel* m, const mjData* d, int id,
#                             mjtNum distance, mjtNum* constimp);
#
# # solver reference
# typedef void (*mjfSolRef)(const mjModel* m, const mjData* d, int id,
#                           mjtNum constimp, mjtNum imp, int dim, mjtNum* ref);
#
# # collision detection
# typedef int (*mjfCollision)(const mjModel* m, const mjData* d,
#                             mjContact* con, int g1, int g2, mjtNum margin);



cdef extern from "mjrender.h" nogil:

    ctypedef enum mjtGridPos:            # grid position for overlay
        mjGRID_TOPLEFT      = 0,        # top left
        mjGRID_TOPRIGHT,                # top right
        mjGRID_BOTTOMLEFT,              # bottom left
        mjGRID_BOTTOMRIGHT              # bottom right


    ctypedef enum mjtFramebuffer:        # OpenGL framebuffer option
        mjFB_WINDOW         = 0,        # default/window buffer
        mjFB_OFFSCREEN                  # offscreen buffer


    ctypedef enum mjtFontScale:          # font scale, used at context creation
        mjFONTSCALE_100     = 100,      # normal scale, suitable in the absence of DPI scaling
        mjFONTSCALE_150     = 150,      # 150% scale
        mjFONTSCALE_200     = 200       # 200% scale


    ctypedef enum mjtFont:               # font type, used at each text operation
        mjFONT_NORMAL       = 0,        # normal font
        mjFONT_SHADOW,                  # normal font with shadow (for higher contrast)
        mjFONT_BIG                      # big font (for user alerts)

    ctypedef struct mjrRect:                     # OpenGL rectangle
        int left                        # left (usually 0)
        int bottom                      # bottom (usually 0)
        int width                       # width (usually buffer width)
        int height                      # height (usually buffer height)


    ctypedef struct mjrContext:                  # custom OpenGL context
        # parameters copied from mjVisual
        float lineWidth                 # line width for wireframe rendering
        float shadowClip                # clipping radius for directional lights
        float shadowScale               # fraction of light cutoff for spot lights
        int shadowSize                  # size of shadow map texture
        int offWidth                    # width of offscreen buffer
        int offHeight                   # height of offscreen buffer
        int offSamples                  # number of offscreen buffer multisamples

        # offscreen rendering objects
        unsigned int offFBO             # offscreen framebuffer object
        unsigned int offFBO_r           # offscreen framebuffer for resolving multisamples
        unsigned int offColor           # offscreen color buffer
        unsigned int offColor_r         # offscreen color buffer for resolving multisamples
        unsigned int offDepthStencil    # offscreen depth and stencil buffer
        unsigned int offDepthStencil_r  # offscreen depth and stencil buffer for resolving multisamples

        # shadow rendering objects
        unsigned int shadowFBO          # shadow map framebuffer object
        unsigned int shadowTex          # shadow map texture

        # texture objects and info
        int ntexture                    # number of allocated textures
        int textureType[100]            # type of texture (mjtTexture)
        unsigned int texture[100]       # texture names

        # displaylist starting positions
        unsigned int basePlane          # all planes from model
        unsigned int baseMesh           # all meshes from model
        unsigned int baseHField         # all hfields from model
        unsigned int baseBuiltin        # all buildin geoms, with quality from model
        unsigned int baseFontNormal     # normal font
        unsigned int baseFontShadow     # shadow font
        unsigned int baseFontBig        # big font

        # displaylist ranges
        int     rangePlane              # all planes from model
        int     rangeMesh               # all meshes from model
        int     rangeHField             # all hfields from model
        int     rangeBuiltin            # all builtin geoms, with quality from model
        int     rangeFont               # all characters in font

        # character info
        int     charWidth[127]          # character widths: normal and shadow
        int     charWidthBig[127]       # chacarter widths: big
        int     charHeight              # character heights: normal and shadow
        int     charHeightBig           # character heights: big

        # capabilities
        int     glewInitialized         # is glew initialized
        int     windowAvailable         # is default/window framebuffer available
        int     windowSamples           # number of samples for default/window framebuffer
        int     windowStereo            # is stereo available for default/window framebuffer
        int     windowDoublebuffer      # is default/window framebuffer double buffered

        # only field that changes after mjr_makeContext
        int     currentBuffer           # currently active framebuffer: mjFB_WINDOW or mjFB_OFFSCREEN




cdef class MjSim(object):
    """
    Wrapper around PyMjModel and PyMjData with extra functionality for
    advancing the simulation dynamics and keeping a buffer between steps.

    The XML used to create the internal model is accessible if the model was
    created by load_model_from_xml(), and is at `self.model.xml`.

    MuJoCo can serialize a model to XML for you, and this is done using the
    save_model_to_xml() method.  For example:
        save_model_to_xml(self.model)
    """
    # MjRenderContext for rendering camera views.
    cdef readonly list render_contexts
    cdef readonly object _render_context_window
    cdef readonly object _render_context_offscreen

    # MuJoCo model
    cdef readonly PyMjModel model
    # MuJoCo data
    cdef readonly PyMjData data
    # Number of substeps when calling .step
    cdef readonly int nsubsteps
    # User defined state.
    cdef readonly dict udd_state
    # User defined dynamics callback
    cdef readonly object _udd_callback
    # Allows to store extra information in MjSim.
    cdef readonly dict extras

    def __cinit__(self, PyMjModel model, PyMjData data=None, int nsubsteps=1,
                  udd_callback=None):
        """
        Args:
            - model (PyMjModel): model to simulate.
            - data (PyMjData): stores the simulation data. Will be created
                if not provided.
            - nsubsteps (int): number of substeps to run on `.step`. Buffers
                will be swapped only once per step.
            - udd_callback: callback for the user-defined dynamics. It
                takes an MjSim object as the current state and returns the
                next udd_state after applying dynamics. Useful for reward
                functions that operate over functions of historical state.
        """
        self.nsubsteps = nsubsteps
        self.model = model
        if data is None:
            with wrap_mujoco_warning():
                _data = mj_makeData(self.model.ptr)
            if _data == NULL:
                raise Exception('mj_makeData failed!')
            self.data = WrapMjData(_data, self.model)
        else:
            self.data = data

        self.render_contexts = []
        self._render_context_offscreen = None
        self._render_context_window = None
        self.udd_state = None
        self.udd_callback = udd_callback
        self.extras = {}

    def reset(self):
        """
        Resets the simulation data and clears buffers.
        """
        with wrap_mujoco_warning():
            mj_resetData(self.model.ptr, self.data.ptr)

        self.udd_state = None
        self.step_udd()

    def forward(self):
        """
        Calls `mj_forward`.
        """
        with wrap_mujoco_warning():
            mj_forward(self.model.ptr, self.data.ptr)

    def step(self):
        """
        Calls `mj_step`, with `nsubsteps` as specified on `MjSim` instance creation.

        If qpos or qvel have been modified directly, the user is required to call
        forward before step if their udd_callback requires access to Mujoco state
        set during the forward dynamics.
        """
        self.step_udd()

        with wrap_mujoco_warning():
            for _ in range(self.nsubsteps):
                mj_step(self.model.ptr, self.data.ptr)

    def render(self, width=None, height=None, *, camera_name=None, depth=False,
               mode='offscreen'):
        """
        Renders view from a camera and returns image as an `numpy.ndarray`.

        Args:
        - width (int): desired image width.
        - height (int): desired image height.
        - camera_name (str): name of camera in model. If None, the free
            camera will be used.
        - depth (bool): if True, also return depth buffer

        Returns:
        - rgb (uint8 array): image buffer from camera
        - depth (float array): depth buffer from camera (only returned
            if depth=True)
        """
        if camera_name is None:
            camera_id = None
        else:
            camera_id = self.model.camera_name2id(camera_name)

        if mode == 'offscreen':
            if self._render_context_offscreen is None:
                render_context = MjRenderContextOffscreen(self)
            else:
                render_context = self._render_context_offscreen

            render_context.render(
                width=width, height=height, camera_id=camera_id)
            return render_context.read_pixels(
                width, height, depth=depth)
        elif mode == 'window':
            if self._render_context_window is None:
                from pymj.mjviewer import MjViewer
                render_context = MjViewer(self)
            else:
                render_context = self._render_context_window

            render_context.render()

        else:
            raise ValueError("Mode must be either 'window' or 'offscreen'.")

    def add_render_context(self, render_context):
        self.render_contexts.append(render_context)
        if render_context.offscreen and self._render_context_offscreen is None:
            self._render_context_offscreen = render_context
        elif not render_context.offscreen and self._render_context_window is None:
            self._render_context_window = render_context

    @property
    def udd_callback(self):
        return self._udd_callback

    @udd_callback.setter
    def udd_callback(self, value):
        self._udd_callback = value
        self.udd_state = None
        self.step_udd()

    def step_udd(self):
        if self._udd_callback is None:
            self.udd_state = {}
        else:
            schema_example = self.udd_state
            self.udd_state = self._udd_callback(self)
            # Check to make sure the udd_state has consistent keys and dimension across steps
            if schema_example is not None:
                keys = set(schema_example.keys()) | set(self.udd_state.keys())
                for key in keys:
                    assert key in schema_example, "Keys cannot be added to udd_state between steps."
                    assert key in self.udd_state, "Keys cannot be dropped from udd_state between steps."
                    if isinstance(schema_example[key], Number):
                        assert isinstance(self.udd_state[key], Number), \
                            "Every value in udd_state must be either a number or a numpy array"
                    else:
                        assert isinstance(self.udd_state[key], np.ndarray), \
                            "Every value in udd_state must be either a number or a numpy array"
                        assert self.udd_state[key].shape == schema_example[key].shape, \
                            "Numpy array values in udd_state must keep the same dimension across steps."

    def get_state(self):
        """ Returns a copy of the simulator state. """
        return MjSimState(np.copy(self.data.qpos),
                          np.copy(self.data.qvel),
                          copy.deepcopy(self.udd_state))

    def set_state(self, value):
        """
        Sets the state from an MjSimState.
        If the MjSimState was previously unflattened from a numpy array, consider
        set_state_from_flattened, as the defensive copy is a substantial overhead
        in an inner loop.

        Args:
        - value (MjSimState): the desired state.
        - call_forward: optionally call sim.forward(). Called by default if
            the udd_callback is set.
        """
        self.data.qpos[:] = np.copy(value.qpos)
        self.data.qvel[:] = np.copy(value.qvel)
        self.udd_state = copy.deepcopy(value.udd_state)

    def set_state_from_flattened(self, value):
        """ This helper method sets the state from an array without requiring a defensive copy."""
        state = MjSimState.from_flattened(value, self)

        self.data.qpos[:] = state.qpos
        self.data.qvel[:] = state.qvel
        self.udd_state = state.udd_state
